#include "config.h"
#include "libvmma.h"
#include "buddy.h"
#include "os_dep.h"
#include <string.h>

#define CTX_ALIGN 64

#if VMMA_FEAT_MULTITHREADED
#  define VMMA_LOCK(_ctx_, _ret_) \
		{ \
			const unsigned int do_lock = (VMMA_FL_MULTITHREADED & (_ctx_)->flags); \
			if(do_lock && 0!=vmma_semwait((_ctx_)->sema, 1)) { return _ret_; }
#  define VMMA_UNLOCK(_ctx_) \
			if(do_lock) { vmma_semsig((_ctx_)->sema); } \
		}
#else /* VMMA_FEAT_MULTITHREADED */
#  define VMMA_LOCK(...)
#  define VMMA_UNLOCK(...)
#endif /* VMMA_FEAT_MULTITHREADED */

struct clientnode;

struct clientnode
{
	struct clientnode* next;
	struct clientnode* prev;
};

struct vmma_context
{
	void*             sema;        /* context semaphore (only if VMMA_FL_MULTITHREADED) */
	unsigned int      flags;       /* accepted flags when created the context */
	struct buddy*     buddy_blk;   /* buddy allocator of the blocks */
	size_t            blocksize;   /* minimal block size to allocate in clients */
	size_t            clientalign; /* original client align */
	size_t            memsize;     /* size of the memory that 'buddy_blk' has */
	size_t            cli_alloc;   /* initial block size of the client */
	size_t            cli_size;    /* size of the actual client (including the block cache) */
	struct clientnode list;        /* list of all created clients */
};

struct vmma_client
{
	struct clientnode    node;  /* link in the vmma_context.list */
	struct vmma_context* ctx;   /* the context the client was made in */
	struct buddy*        buddy; /* the extendable buddy allocator constructed for the client */
};

unsigned int vmma_get_supported_flags(void)
{
	return

#if VMMA_FEAT_MULTITHREADED
	VMMA_FL_MULTITHREADED |
#endif /* VMMA_FEAT_MULTITHREADED */

#if VMMA_FEAT_MINGROWTH
	VMMA_FL_MINGROWTH |
#endif /* VMMA_FEAT_MINGROWTH */

#if VMMA_FEAT_KEEPCACHED
	VMMA_FL_KEEPCACHED |
#endif /* VMMA_FEAT_KEEPCACHED */

	0;
}

/* TODO:
int vmma_get_patch_level(void)
{
	return the max of the revision numbers
}
*/

static void list_init(struct clientnode* const list)
{
	list->next = list->prev = list;
}

static struct clientnode* list_first(struct clientnode* const list)
{
	return list->next;
}

static struct clientnode* list_next(struct clientnode* const node)
{
	return node->next;
}

static void list_remove(struct clientnode* const node)
{
	node->prev->next = node->next;
	node->next->prev = node->prev;
}

static void list_insert_tail(struct clientnode* const list, struct clientnode* const node)
{
	node->next = list;
	node->prev = list->prev;
	list->prev->next = node;
	list->prev = node;
}

static struct vmma_context* get_aligned_context(void* const memory)
{
	vmma_uintptr_t addr = (vmma_uintptr_t)memory;
	addr += CTX_ALIGN - 1;
	addr &= ~(((vmma_uintptr_t)CTX_ALIGN) - 1);
	return (struct vmma_context*)addr;
}

static size_t mem_diff(void* const base, void* const further)
{
	const vmma_uintptr_t B = (vmma_uintptr_t)base;
	const vmma_uintptr_t F = (vmma_uintptr_t)further;
	return (size_t)(F - B);
}

struct vmma_context* vmma_create_context(void* const memory, const size_t memsize,
	const size_t clientalign, const size_t blocksize, const unsigned int flags)
{
	struct vmma_context* ctx;

	if(NULL==memory)
	{
		return NULL;
	}

	if(CTX_ALIGN>memsize)
	{
		return NULL;
	}

	if(VMMA_BLOCKSIZE_MIN>blocksize)
	{
		return NULL;
	}

	if(VMMA_CLIENTALIGN_MIN>clientalign)
	{
		return NULL;
	}

	if(0==memsize || memsize<(sizeof(struct vmma_context)+blocksize*2))
	{
		return NULL;
	}

	if(clientalign & (clientalign-1))
	{
		return NULL;
	}

	if(blocksize & (blocksize-1))
	{
		return NULL;
	}

	if(clientalign>blocksize)
	{
		return NULL;
	}

	ctx = get_aligned_context(memory);
	memset(ctx, 0, sizeof(*ctx));

	list_init(&ctx->list);

	if(NULL==(ctx->buddy_blk=buddy_init(ctx+1, memsize-mem_diff(memory, ctx+1), blocksize)))
	{
		return NULL;
	}

	ctx->memsize = buddy_get_total_size(ctx->buddy_blk);

#if VMMA_FEAT_MULTITHREADED
	if(VMMA_FL_MULTITHREADED & flags)
	{
		if(NULL==(ctx->sema=vmma_semget(1)))
		{
			return NULL;
		}

		ctx->flags |= VMMA_FL_MULTITHREADED;
	}
#endif /* VMMA_FEAT_MULTITHREADED */

#if VMMA_FEAT_MINGROWTH
	if(VMMA_FL_MINGROWTH & flags) { ctx->flags |= VMMA_FL_MINGROWTH; }
#endif /* VMMA_FEAT_MINGROWTH */

#if VMMA_FEAT_KEEPCACHED
	if(VMMA_FL_KEEPCACHED & flags) { ctx->flags |= VMMA_FL_KEEPCACHED; }
#endif /* VMMA_FEAT_KEEPCACHED */

	ctx->clientalign = clientalign;
	ctx->blocksize   = blocksize;
	ctx->cli_alloc   = ctx->blocksize;
	ctx->cli_size    = sizeof(struct vmma_client);

	return ctx;
}

void vmma_delete_context(struct vmma_context* const ctx)
{
	if(VMMA_FL_MULTITHREADED & ctx->flags)
	{
		vmma_semdel(ctx->sema);
	}

	/* we may have extra memory blocks allocated from the system */
	buddy_deinit(ctx->buddy_blk);

	memset(ctx, 0, sizeof(*ctx));
}

unsigned int vmma_get_accepted_flags(struct vmma_context* const ctx)
{
	return ctx->flags;
}

size_t vmma_get_total_size(struct vmma_context* const ctx)
{
	return ctx->memsize;
}

static size_t vmma_get_free_size_unlocked(struct vmma_context* const ctx)
{
	return buddy_get_total_free(ctx->buddy_blk);
}

size_t vmma_get_free_size(struct vmma_context* const ctx)
{
	size_t res;
	VMMA_LOCK(ctx, 0);
	res = vmma_get_free_size_unlocked(ctx);
	VMMA_UNLOCK(ctx);
	return res;
}

static size_t vmma_populate_free_size_unlocked(struct vmma_context* const ctx)
{
	size_t             res  = buddy_get_total_free(ctx->buddy_blk);
	struct clientnode* node = list_first(&ctx->list);

	while(node!=(&ctx->list))
	{
		struct vmma_client* const client = (struct vmma_client*)node;

		res += buddy_get_total_free(client->buddy);

		node = list_next(node);
	}

	return res;
}

size_t vmma_populate_free_size(struct vmma_context* const ctx)
{
	size_t res;
	VMMA_LOCK(ctx, 0);
	res = vmma_populate_free_size_unlocked(ctx);
	VMMA_UNLOCK(ctx);
	return res;
}

static void* wrap_buddy_alloc(void* b, size_t n)
{
	return buddy_alloc((struct buddy*)b, n);
}

static void wrap_buddy_free(void* b, void* m)
{
	buddy_free((struct buddy*)b, m);
}

#define FASTGROW_EXP_FLAGS ((unsigned int)(BUDDY_EXP_INCREMENTAL | BUDDY_EXP_PREALLOC))

static struct vmma_client* vmma_create_client_unlocked(struct vmma_context* const ctx, const size_t limit)
{
	const size_t              size    = ctx->cli_alloc;
	struct vmma_client* const client  = (struct vmma_client*)buddy_alloc(ctx->buddy_blk, size);
	size_t                    max_exp = 14; /* TODO: this is a random value */
	unsigned int              exp_fl  = BUDDY_EXP_ALLOCPOW2 | FASTGROW_EXP_FLAGS;

	if(NULL==client)
	{
		return NULL;
	}

	memset(client, 0, ctx->cli_size);

	/* TODO: use the 'limit' parameter! */

#if VMMA_FEAT_MINGROWTH
	if(VMMA_FL_MINGROWTH & ctx->flags)
	{
		exp_fl  &= ~FASTGROW_EXP_FLAGS;
		max_exp += 10; /* TODO: this is a random value */
	}
#endif /* VMMA_FEAT_MINGROWTH */

	if(NULL==(client->buddy=buddy_init_extendable(
		((vmma_uint8_t*)client)+ctx->cli_size, size-ctx->cli_size,
		ctx->clientalign,
		max_exp,
		exp_fl,
		ctx->buddy_blk,
		wrap_buddy_alloc,
		wrap_buddy_free)))
	{
		buddy_free(ctx->buddy_blk, client);
		return NULL;
	}

	client->ctx = ctx;

	list_insert_tail(&ctx->list, &client->node);

	return client;
}

struct vmma_client* vmma_create_client(struct vmma_context* const ctx, const size_t limit)
{
	struct vmma_client* client;
	VMMA_LOCK(ctx, 0);
	client = vmma_create_client_unlocked(ctx, limit);
	VMMA_UNLOCK(ctx);
	return client;
}

static void vmma_delete_client_unlocked(struct vmma_client* const client)
{
	list_remove(&client->node);
	buddy_deinit(client->buddy);
	buddy_free(client->ctx->buddy_blk, client);
}

void vmma_delete_client(struct vmma_client* const client)
{
	struct vmma_context* const ctx = client->ctx;

	VMMA_LOCK(ctx, ;);
	vmma_delete_client_unlocked(client);
	VMMA_UNLOCK(ctx);
}

static void* vmma_alloc_unlocked(struct vmma_client* const client, size_t size)
{
	return buddy_alloc(client->buddy, size);
}

void* vmma_alloc(struct vmma_client* const client, size_t size)
{
	struct vmma_context* const ctx = client->ctx;

	void* res;
	VMMA_LOCK(ctx, NULL);
	res = vmma_alloc_unlocked(client, size);
	VMMA_UNLOCK(ctx);
	return res;
}

static void vmma_free_unlocked(struct vmma_client* const client, void* memory)
{
	buddy_free(client->buddy, memory);
}

void vmma_free(struct vmma_client* const client, void* memory)
{
	struct vmma_context* const ctx = client->ctx;

	VMMA_LOCK(ctx, ;);
	vmma_free_unlocked(client, memory);
	VMMA_UNLOCK(ctx);
}

static size_t vmma_get_size_unlocked(struct vmma_client* const client, void* memory)
{
	return buddy_msize(client->buddy, memory);
}

size_t vmma_get_size(struct vmma_client* const client, void* memory)
{
	struct vmma_context* const ctx = client->ctx;

	size_t res;
	VMMA_LOCK(ctx, 0);
	res = vmma_get_size_unlocked(client, memory);
	VMMA_UNLOCK(ctx);
	return res;
}

static void* vmma_test_address_unlocked(struct vmma_client* const client, void* address)
{
	if(0<buddy_msize(client->buddy, address))
	{
		return address;
	}

	return NULL;
}

void* vmma_test_address(struct vmma_client* const client, void* address)
{
	struct vmma_context* const ctx = client->ctx;

	void* res;
	VMMA_LOCK(ctx, 0);
	res = vmma_test_address_unlocked(client, address);
	VMMA_UNLOCK(ctx);
	return res;
}

size_t vmma_get_client_limit(struct vmma_client* const client)
{
	/* ---- TODO ---- */
	/* Query the client memory usage limit specified in vmma_create_client(). */
	return 0;
}

static size_t vmma_get_client_working_size_unlocked(struct vmma_client* const client)
{
	return buddy_get_total_size(client->buddy);
}

size_t vmma_get_client_working_size(struct vmma_client* const client)
{
	struct vmma_context* const ctx = client->ctx;

	size_t res;
	VMMA_LOCK(ctx, 0);
	res = vmma_get_client_working_size_unlocked(client);
	VMMA_UNLOCK(ctx);
	return res;
}

static size_t vmma_get_client_usage_unlocked(struct vmma_client* const client)
{
	return buddy_get_total_size(client->buddy) - buddy_get_total_free(client->buddy);
}

size_t vmma_get_client_usage(struct vmma_client* const client)
{
	struct vmma_context* const ctx = client->ctx;

	size_t res;
	VMMA_LOCK(ctx, 0);
	res = vmma_get_client_usage_unlocked(client);
	VMMA_UNLOCK(ctx);
	return res;
}

size_t vmma_populate_cached_blocks(struct vmma_client* const client)
{
	/* ---- TODO ---- */
	/* Return the total size of cached memory blocks. */
	/* Not efficient, shall not be called often. */
	return 0;
}

void vmma_purge_client_cache(struct vmma_client* const client)
{
	/* ---- TODO ---- */
	/* Force to purge cached memory blocks. This may reduce internal */ 
	/* fragmentation, and makes a larger allocation possible. */
	/* Not efficient, shall not be called often. */
	return;
}
