#include "config.h"
#include "test.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <assert.h>
#include "libvmma/buddy.h"

#define DO_SHUFFLE     1 /* shall be enabled */
#define DISPLAY_INFO   0 /* shall be disabled if "usually works" */
#define DISPLAY_ROUNDS 0 /* shall be disabled if "usually works" */
#define DISPLAY_DOTS   0 /* shall be disabled if "usually works" */
#define DISPLAY_ALLOC  0 /* shall be disabled if "usually works" */
#define DISPLAY_FREE   0 /* shall be disabled if "usually works" */
#define DISPLAY_EBA    0 /* display extendable buddy backend allocations; shall be disabled if "usually works" */

static uint8_t memory[4*1024];
static uint8_t memory2[2*sizeof(memory)];
static int alloc_state;

static void* eb_alloc(void* ctx, size_t size)
{
	if(0<alloc_state) { return NULL; }
	if(size>sizeof(memory2)) { return NULL; }
#if DISPLAY_EBA
	printf("+\n");
#endif /* DISPLAY_EBA */
	++alloc_state;
	return memory2;
}

static void eb_free(void* ctx, void* mem)
{
	if(0<alloc_state && mem==(void*)memory2)
	{
#if DISPLAY_EBA
		printf("-\n");
#endif /* DISPLAY_EBA */
		alloc_state = 0;
	}
}

#if DISPLAY_INFO

static void show_info(struct buddy* const buddy, const char* const title)
{
	printf("======== %s ========\n", title);
	printf("buddy at: %p\n", buddy);
	printf("small:    %lu\n", (unsigned long)buddy_get_smallest_size(buddy));
	printf("large:    %lu\n", (unsigned long)buddy_get_largest_size(buddy));
	printf("largest:  %lu\n", (unsigned long)buddy_get_largest_free(buddy));
	printf("total:    %lu\n", (unsigned long)buddy_get_total_size(buddy));
	printf("free:     %lu\n", (unsigned long)buddy_get_total_free(buddy));
	printf("========================\n");
}

#else /* DISPLAY_INFO */
#define show_info(...)
#endif /* DISPLAY_INFO */

#if DO_SHUFFLE

#include "shuffle.h"

static void shuffle(void** array, size_t nelem)
{
	shuffle_pointers(array, nelem);
}

#endif /* DO_SHUFFLE */

static struct buddy* create_buddy(const int use_ebuddy)
{
	const size_t minsize = 32;

	if(!use_ebuddy)
	{
		return buddy_init(memory, sizeof(memory), minsize);
	}

	alloc_state = 0;

	return buddy_init_extendable(memory, sizeof(memory), minsize,
		2, 0, NULL, eb_alloc, eb_free);
}

int test_buddy(const int use_ebuddy, const unsigned int nrounds)
{
	struct buddy* const buddy = create_buddy(use_ebuddy);
	size_t small;
	size_t num;
	size_t i;
	size_t round;
	void** ptrs;

	if(NULL==buddy)
	{
		fprintf(stderr, "failed to init buddy\n");
		return 1;
	}

	show_info(buddy, "before");

	small = buddy_get_smallest_size(buddy);
	num   = (buddy_get_total_free(buddy) / small) * (use_ebuddy ? 3 : 1);
	ptrs  = (void**)calloc(num, sizeof(void*));

	if(NULL==ptrs)
	{
		fprintf(stderr, "out of system memory\n");
		return 1;
	}

	for(round=1;round<=nrounds;++round)
	{
		const size_t nfree = buddy_get_total_free(buddy);

#if DISPLAY_ROUNDS
		printf("Round %lu\n", (unsigned long)round);
#elif DISPLAY_DOTS
		fputc('.', stdout);
		fflush(stdout);
#endif /* DISPLAY_DOTS */

		for(i=0;i<num;++i)
		{
			void* const ptr = buddy_alloc(buddy, small);
			assert(NULL!=ptr);
#if !DISPLAY_DOTS && DISPLAY_ALLOC
			printf("  alloc #%lu: %p\n", (unsigned long)i, ptr);
#endif /* !DISPLAY_DOTS && DISPLAY_ALLOC */
			ptrs[i] = ptr;
		}

		if(!use_ebuddy)
		{
			assert(NULL==buddy_alloc(buddy, 1));
		}

#if DO_SHUFFLE
		srand(100+round);
		shuffle(ptrs, num);
#endif /* DO_SHUFFLE */

		for(i=0;i<num;++i)
		{
#if !DISPLAY_DOTS && DISPLAY_FREE
			printf("  free #%lu: %p\n", (unsigned long)i, ptrs[i]);
#endif /* !DISPLAY_DOTS && DISPLAY_FREE */
			buddy_free(buddy, ptrs[i]);
		}

		if(!use_ebuddy || 1<round)
		{
			assert(buddy_get_total_free(buddy)==nfree);
		}
	}

#if !DISPLAY_ROUNDS && DISPLAY_DOTS
	fputc('\n', stdout);
#endif /* DISPLAY_DOTS */

	show_info(buddy, "after");
	buddy_deinit(buddy);
	free(ptrs);

	return 0;
}
