#include "config.h"
#include "test.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <assert.h>
#include "libvmma/libvmma.h"

#define DO_SHUFFLE     1 /* shall be enabled */
#define DISPLAY_INFO   0 /* shall be disabled if "usually works" */
#define DISPLAY_ROUNDS 0 /* shall be disabled if "usually works" */
#define DISPLAY_DOTS   0 /* shall be disabled if "usually works" */
#define DISPLAY_ALLOC  0 /* shall be disabled if "usually works" */
#define DISPLAY_FREE   0 /* shall be disabled if "usually works" */

static uint8_t memory[128*1024];

#if DISPLAY_INFO

static void show_ctx_info(struct vmma_context* const ctx, const char* const title)
{
	printf("======== %s ========\n", title);
	printf("context at: %p\n", ctx);
	printf("size:       %lu\n", (unsigned long)vmma_get_total_size(ctx));
	printf("free:       %lu\n", (unsigned long)vmma_get_free_size(ctx));
	printf("free+:      %lu\n", (unsigned long)vmma_populate_free_size(ctx));
	printf("========================\n");
}

static void show_client_info(struct vmma_client* const client, const char* const title)
{
	printf("======== %s ========\n", title);
	printf("client at: %p\n", client);
	printf("working:   %lu\n", (unsigned long)vmma_get_client_working_size(client));
	printf("usage:     %lu\n", (unsigned long)vmma_get_client_usage(client));
	printf("cache:     %lu\n", (unsigned long)vmma_populate_cached_blocks(client));
	printf("========================\n");
}

#else /* DISPLAY_INFO */
#define show_ctx_info(...)
#define show_client_info(...)
#endif /* DISPLAY_INFO */

#if DO_SHUFFLE

#include "shuffle.h"

static void shuffle(void** array, size_t nelem)
{
	shuffle_pointers(array, nelem);
}

#endif /* DO_SHUFFLE */

#define MAX_CLI 2
#define BLOCK   4096

int test_vmma(const unsigned int flags, const unsigned int nclients, const unsigned int nrounds)
{
	struct vmma_context* const ctx = vmma_create_context(memory, sizeof(memory),
		VMMA_CLIENTALIGN_MIN,
		BLOCK,
		flags);

	const size_t alloc_max = (VMMA_FL_MINGROWTH & (flags & vmma_get_supported_flags()))
		?
		(((sizeof(memory) - BLOCK*(nclients*2+1)) / VMMA_CLIENTALIGN_MIN) / (2 + nclients))
		:
		((BLOCK*2/nclients) / VMMA_CLIENTALIGN_MIN)
		;

	struct vmma_client* clients[MAX_CLI];

	size_t num[MAX_CLI];
	size_t i;
	size_t round;
	void** ptrs[MAX_CLI];

	if(NULL==ctx)
	{
		fprintf(stderr, "failed to create context\n");
		return 1;
	}

	if(1>nclients || MAX_CLI<nclients)
	{
		fprintf(stderr, "'nclients==%u' is not supported\n", nclients);
		return 1;
	}

	for(i=0;i<nclients;++i)
	{
		if(NULL==(clients[i]=vmma_create_client(ctx, 0)))
		{
			fprintf(stderr, "failed to create client #%lu\n", (unsigned long)(i+1));
			return 1;
		}

		show_client_info(clients[i], "before");
		num[i] = 0;
	}

	show_ctx_info(ctx, "before");

	ptrs[0] = (void**)calloc(alloc_max*nclients, sizeof(void*));

	if(NULL==ptrs[0])
	{
		fprintf(stderr, "out of system memory\n");
		return 1;
	}

	for(i=1;i<nclients;++i)
	{
		ptrs[i] = ptrs[i-1] + alloc_max;
	}

	for(round=1;round<=nrounds;++round)
	{
#if DISPLAY_ROUNDS
		printf("Round %lu\n", (unsigned long)round);
#elif DISPLAY_DOTS
		fputc('.', stdout);
		fflush(stdout);
#endif /* DISPLAY_DOTS */

		for(i=0;i<alloc_max;++i)
		{
			const size_t cli = (round+i)%nclients;

			void* const ptr = vmma_alloc(clients[cli], VMMA_CLIENTALIGN_MIN);
			assert(NULL!=ptr);
#if !DISPLAY_DOTS && DISPLAY_ALLOC
			printf("  alloc #%lu of %lu (cli %lu): %p\n", (unsigned long)i, (unsigned long)alloc_max, (unsigned long)cli, ptr);
#endif /* !DISPLAY_DOTS && DISPLAY_ALLOC */
			ptrs[cli][num[cli]++] = ptr;
		}

#if DO_SHUFFLE
		srand(100+round);
		for(i=0;i<nclients;++i)
		{
			shuffle(ptrs[i], num[i]);
		}
#endif /* DO_SHUFFLE */

		for(i=0;i<nclients;++i)
		{
			while(0<num[i])
			{
				void* const ptr = ptrs[i][num[i]-1];
#if !DISPLAY_DOTS && DISPLAY_FREE
				printf("  free #%lu (cli %lu): %p\n", (unsigned long)num[i], (unsigned long)i, ptr);
#endif /* !DISPLAY_DOTS && DISPLAY_FREE */
				--num[i];
				vmma_free(clients[i], ptr);
			}
		}
	}

#if !DISPLAY_ROUNDS && DISPLAY_DOTS
	fputc('\n', stdout);
#endif /* DISPLAY_DOTS */

	for(i=0;i<nclients;++i)
	{
		show_client_info(clients[i], "after");
		vmma_delete_client(clients[i]);
	}

	show_ctx_info(ctx, "after");
	vmma_delete_context(ctx);
	free(ptrs[0]);

	return 0;
}
